var pg = require('pg');
require("pg-parse-float")(pg);
// var pgCursor = require('pg-cursor');
var PgQueryStream = require('pg-query-stream');

var mg = require('mongodb');
var MgClient = mg.MongoClient;
var MgServer = mg.Server;

var DB_TYPE_MONGODB    = 'mongodb';
var DB_TYPE_POSTGRESQL = 'postgresql';

function Database(config) {
  if (!config.type) throw new Error('Database type is required');
  if (!config.name) throw new Error('Database name is required');
  if (!config.host) throw new Error('Database host is required');
  if (!config.port) throw new Error('Database port is required');

  // standardize config.type
  switch (config.type) {
    case 'pg':
    case 'postgres':
    case 'postgresql':
      config.type = DB_TYPE_POSTGRESQL;
      break;
    case 'mg':
    case 'mongo':
    case 'mongodb':
      config.type = DB_TYPE_MONGODB;
      break;
    default:
      throw new Error('Unknown database type: ' + config.type);
      break;
  }

  this.client = null;
  this.debug = config.debug;

  this.type = config.type;
  this.name = config.name;
  this.host = config.host;
  this.port = config.port;
  this.user = config.user;
  this.pass = config.pass;

  // automatically generate the alias if not defined in config
  this.alias = config.alias || config.type + '/' + config.name;
}

Database.prototype.connect = function (callback) {
  var self = this;

  if (self.connecting) { // connection in progress, try again
    return setTimeout(self.connect.bind(self, callback), 100);
  } else if (self.client && self.connected) {
    return process.nextTick(function () {
      callback(null, self.client);
    });
  }

  self.connecting = true;
  self.connected = false;
  if (self.debug) console.info('Connecting to ' + self.alias +
    ' at ' + self.host + ':' + self.port);

  switch (self.type) {
    case DB_TYPE_MONGODB:
      var url = "mongodb://" + self.host + ":" + self.port + "/" + self.name;
      var mgServer = new MgServer(self.host, self.port, {
        auto_reconnect: true
      });
      var mgClient = new MgClient(mgServer, {
        w: 1,
        journal: true
      });

      mgClient.open(function(err, client) {
        self.connecting = false;

        if (err) {
          return mgClient.close(callback.bind(self, err));
        } else if (client) {
          var db = client.db(self.name);
          self.client = db;
          self.connected = true;
          if (self.debug) console.info('Connected to ' + self.alias +
            ' at ' + self.host + ':' + self.port);
          return callback(err, db);
        }
      });
      break;
    case DB_TYPE_POSTGRESQL:
      var url = "postgres://" +
        (self.user ? self.user + (self.pass ? ":" + self.pass : "") + "@" : "") +
        (self.host + (self.port ? ":" + self.port : "") + "/" + self.name);

      return pg.connect(url, function(err, client, done) {
        self.connecting = false;

        if (client) {
          self.client = client;
          self.connected = true;
          if (self.debug) console.info('Connected to ' + self.alias +
            ' at ' + self.host + ':' + self.port);

          client.close = function (callback) {
            done();
            callback();
          };

          client.stream = function (query, options, callback) {
            // options object is optional
            if (arguments.length === 2) {
              callback = options;
              options = undefined;
            }

            // allow query as a string if values array is not required
            if (typeof query === 'string') query = { text: query, values: [] };

            // execute query and create readable stream
            var qs = new PgQueryStream(query.text, query.values, options);
            var stream = client.query(qs);

            // use callback function if provided
            if (callback) process.nextTick(callback.bind(this, null, stream));

            return stream;
          };

          // TODO: not working yet, waiting for pg-cursor to mature
          // client.cursor = function (query, callback) {
          //   // allow query as a string if values array is not required
          //   if (typeof query === 'string') query = { text: query, values: [] };
          //   var cs = new pgCursor(query.text, query.values);
          //   var cursor = client.query(cs);
          //   // use callback function if provided
          //   if (callback) process.nextTick(callback.bind(this, null, cursor));
          //   return cursor;
          // };
        }

        return callback(err, client);
      });
      break;
  }
};

Database.prototype.close = function (callback) {
  if (this.client && this.connected) {
    return this.client.close(callback);
  } else if (this.client && this.connecting) {
    return setTimeout(this.close.bind(this, callback), 100);
  } else {
    return process.nextTick(callback);
  }
};

module.exports = Database;
